#' self-X autonomic supervised feature selection
#' @description
#' With the aim of making SUPERVISED feature selection more autonomic,
#' an ad-hoc procedure is implemented here following next approach:
#' \enumerate{
#' \item Initially, well-known technique as RELIEF is used
#' to select the most important features under different criteria.
#' \item Subsequently, an OWA operator is applied to to determine which
#'  3 variables are the most relevant to be used as input in the following
#'  modelling phase
#' }
#' @keywords package selfX s-X-AIPI autonomic feature selection
#' @details
#' The feature selection process applied to the data is as follows:
#' \itemize{
#' \item The available data sets are randomly divided into 5 training
#' and test datasets to have a 5-fold cross-validation process.
#' \item Different supervised feature selection techniques are applied to the data
#'   to establish different importance values of the variables considered. First
#'   an evaluator filter method based on RELIEF is used and then 6 different
#'   direct search algorithms (cutt-off method) are applied using a 5-fold
#'   cross validation approach:
#'    \enumerate{
#'     \item  (1) Select K best
#'     \item  (2) Select Percentile
#'     \item  (3) Select threshold
#'     \item  (4) Select threshold range
#'     \item  (5) Select difference
#'     \item  (6) Select slope
#'     }
#'     In all cases, the importance value of the variable is assigned based on
#'     the number of times it has been selected.
#' \item Based on the evaluation obtained, a multi-criteria decision-making
#'   process is carried out based on:
#'     \itemize{
#'        \item OWA (Ordered Weighted Averaging) operator: applying the RIM (Regular
#'        Increasing Monotone) quantifier \eqn{\alpha(r)=r^{\alpha}},
#'        \eqn{r\geq 0} with \eqn{orness=0.4} on the evaluations, an importance is
#'        established for each characteristic, selecting in each case only
#'        the most important ones.
#'     }
#' \item A subset of features is selected and returned.
#' }
#' @param dataDF dataframe containing clean data (no outliers,
#' no missing values, no constant values) for data mining activities.
#' ONLY numerical input variables allowed (that it can be normalized or not).
#' @param varsINPUT vector with the names of the input variables.
#' @param varsOUTPUT vector with the name of the output variable. Only one output variable is allowed.
#' @return text string with the names of the selected input variables
#' @references
#'
#' Kenji Kira and Larry A. Rendell. "A practical approach to feature selection"
#' In Machine Learning Proceedings 1992, pages 249-256. Elsevier, 1992.
#'
#' R.R. Yager. "On ordered weighted averaging aggregation operators in
#' multi-criteria decision making" IEEE Transactions on Systems, Man
#' and Cybernetics, 18:183–190, 1998.
#'
#' F. Aragón-Royón, A. Jiménez-Vílchez, A. Arauzo-Azofra, J.M. Benitez "FSinR:
#' an exhaustive package for feature selection" arXiv e-prints, 2020.
#'
#' @note
#' The s-X-AIPI project has received funding from the European Union's Horizon
#' Europe research and innovation programme under grant agreement Nº 101058715
#' \url{https://s-x-aipi-project.eu/}
#' @seealso \href{https://dicits.ugr.es/software/FSinR/#fsinr-package}{FSinR}
#' for RELIEF methods and \code{\link[agop:owa]{owa}} for OWA.
#' @examplesIf FALSE
#' ## load dataset
#' data(mtcars)
#' ## establish inputs and output
#' varsINPUT <- c("mpg","cyl","disp","hp","drat","wt","qsec","vs","am","gear")
#' varsOUTPUT <- c("carb")
#' ## run relief.owa method
#' relief.owa(mtcars,varsINPUT,varsOUTPUT)
#' @export
relief.owa <- function(dataDF, varsINPUT, varsOUTPUT) {
  watchful <- list(cod = 0, txt = "")

  dataNN <- dataDF
  # COMPROBAR LOS PARAMETROS DE ENTRADA
  if (is.null(dataNN) | nrow(dataNN) == 0 | ncol(dataNN) == 0) {
    watchful$cod <- 1
    watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: data not available", sep = " ")
    watchful$txt <- paste(watchful$txt, "", sep = "\n")
  }
  todas_numericas <- all(sapply(dataNN, is.numeric))
  if (!todas_numericas) {
    watchful$cod <- 1
    watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: not numerical data", sep = " ")
    watchful$txt <- paste(watchful$txt, "", sep = "\n")
  }
  rm(todas_numericas)
  if (is.null(varsINPUT)) {
    watchful$cod <- 1
    watchful$txt <- "ERROR in feature selection function: input variables not available"
  } else {
    for (variable in varsINPUT) {
      if (!variable %in% colnames(dataNN)) {
        watchful$cod <- 1
        watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: input variable", variable, " not presented in the data", sep = " ")
        watchful$txt <- paste(watchful$txt, "", sep = "\n")
      }
    }
  }
  if (is.null(varsOUTPUT)) {
    watchful$cod <- 1
    watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: output variable not available", sep = " ")
    watchful$txt <- paste(watchful$txt, "", sep = "\n")
  } else {
    if (length(varsOUTPUT) > 1) {
      watchful$cod <- 1
      watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: only one output variable can be selected", sep = " ")
      watchful$txt <- paste(watchful$txt, "", sep = "\n")
    } else {
      if (!varsOUTPUT %in% colnames(dataNN)) {
        watchful$cod <- 1
        watchful$txt <- paste(watchful$txt, "ERROR in feature selection function: output variable", varsOUTPUT, " not presented in the data", sep = " ")
        watchful$txt <- paste(watchful$txt, "", sep = "\n")
      }
    }
  }

  if (watchful$cod == 0) {
    # Generamos los 5fcv
    # print("5 FCV creating .... ")
    newdata <- dataNN
    set.seed(42)
    # Randomly shuffle the data
    newdata <- newdata[sample(nrow(newdata)), ]
    # Create 5 equally size folds
    folds <- cut(seq(1, nrow(newdata)), breaks = 5, labels = FALSE)
    # Perform 10 fold cross validation
    training_list <- list()
    testing_list <- list()
    for (fcv in 1:5) {
      # Segement your data by fold using the which() function
      testIndexes <- which(folds == fcv, arr.ind = TRUE)
      test <- newdata[testIndexes, ]
      train <- newdata[-testIndexes, ]
      # Use the test and train data partitions however you desire...
      # generar una lista con los 5FCV que se use posteriormente
      training_list[[fcv]] <- train
      testing_list[[fcv]] <- test
      rm(testIndexes, test, train)
    }
    rm(newdata, fcv, folds)
    # print("5 FCV created")

    numVAR <- length(varsINPUT) # vblesEntrada

    nIndices <- 6 # KBest, Percentile, Threshold, ThresholdRange, Difference, Slope
    nOrnessOWA <- 3 # ornes=0.3,0.4,0.9
    orness <- c(0.3, 0.4, 0.9)
    IndicesRELIEF <- matrix(nrow = numVAR, ncol = (nIndices * 2) + (nOrnessOWA * 2))
    colnames(IndicesRELIEF) <- c(
      "select_KBest", "rank_KBest",
      "select_Percentile", "rank_Percentile",
      "select_Threshold", "rank_Threshold",
      "select_ThresholdRange", "rank_ThresholdRange",
      "select_Difference", "rank_Difference",
      "select_Slope", "rank_Slope",
      "OWAvalue03", "OWApor03", "OWAvalue04", "OWApor04", "OWAvalue09", "OWApor09"
    )
    rownames(IndicesRELIEF) <- colnames(dataNN[-ncol(dataNN)])


    select_KBest <- matrix(ncol = numVAR, 0)
    select_Percentile <- matrix(ncol = numVAR, 0)
    select_Threshold <- matrix(ncol = numVAR, 0)
    select_ThresholdRange <- matrix(ncol = numVAR, 0)
    select_Difference <- matrix(ncol = numVAR, 0)
    select_Slope <- matrix(ncol = numVAR, 0)

    no_cores <- round(parallel::detectCores() * 0.8)
    cl <- parallel::makeCluster(no_cores)
    doParallel::registerDoParallel(cl)

    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    skb_direct_search <- FSinR::selectKBest(k = 3)
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectKBest <- skb_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_KBest <- select_KBest + Lista[[i]]$bestFeatures
    }


    # Select percentile
    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    sp_direct_search <- FSinR::selectPercentile()
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectPercentile <- sp_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_Percentile <- select_Percentile + Lista[[i]]$bestFeatures
    }

    # Select threshold
    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    st_direct_search <- FSinR::selectThreshold()
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectThreshold <- st_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_Threshold <- select_Threshold + Lista[[i]]$bestFeatures
    }

    # Select threshold range
    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    str_direct_search <- FSinR::selectThresholdRange()
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectThresholdRange <- str_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_ThresholdRange <- select_ThresholdRange + Lista[[i]]$bestFeatures
    }


    # Select difference
    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    sd_direct_search <- FSinR::selectDifference()
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectDifference <- sd_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_Difference <- select_Difference + Lista[[i]]$bestFeatures
    }

    # Select slope
    filter_evaluator <- FSinR::filterEvaluator("relief", list(neighbours.count = 4))
    ss_direct_search <- FSinR::selectSlope()
    Lista <- foreach(i = 1:5) %dopar% {
      train <- training_list[[i]]
      res_selectSlope <- ss_direct_search(train, varsOUTPUT, filter_evaluator)
    }
    for (i in 1:5) {
      select_Slope <- select_Slope + Lista[[i]]$bestFeatures
    }

    parallel::stopCluster(cl)
    rm(no_cores, cl)

    ### Asignar rankings: por defecto se asigna 1 al valor min, y N al valor max. Para invertir orden multiplicar valor por -1
    IndicesRELIEF[, 1] <- select_KBest
    IndicesRELIEF[, 2] <- rank(select_KBest)
    IndicesRELIEF[, 3] <- select_Percentile
    IndicesRELIEF[, 4] <- rank(select_Percentile)
    IndicesRELIEF[, 5] <- select_Slope
    IndicesRELIEF[, 6] <- rank(select_Slope)
    IndicesRELIEF[, 7] <- select_Threshold
    IndicesRELIEF[, 8] <- rank(select_Threshold)
    IndicesRELIEF[, 9] <- select_ThresholdRange
    IndicesRELIEF[, 10] <- rank(select_ThresholdRange)
    IndicesRELIEF[, 11] <- select_Difference
    IndicesRELIEF[, 12] <- rank(select_Difference)

    rm(select_KBest, select_Percentile, select_Slope, select_Threshold, select_ThresholdRange, select_Difference)

    ############################# OWA sobre las evaluaciones o sobre los rankings  ############################
    for (j in 1:nOrnessOWA) {
      operadorOR <- orness[j]
      # Calculo Pesos
      alpha <- (1 - operadorOR) / operadorOR
      w <- matrix()
      for (i in 1:nIndices) {
        w[i] <- ((i / nIndices)^alpha) - ((i - 1) / nIndices)^alpha
      }
      rm(i, alpha)

      # Relleno tabla
      for (i in 1:numVAR) {
        ### OWA sobre los evaluaciones
        # IndicesRELIEF[i,(nIndices*2)+(j*2)-1] <- agop::owa(IndicesRELIEF[i,c(1,3,5,7,9,11)], w)
        ### OWA sobre los rankings
        IndicesRELIEF[i, (nIndices * 2) + (j * 2) - 1] <- agop::owa(IndicesRELIEF[i, c(2, 4, 6, 7, 10, 12)], w)
      }
      rm(i)
    }
    rm(j, operadorOR)

    IndicesRELIEF

    # #  calcular % segun owa columna 9 para columna 10
    IndicesRELIEF[, 14] <- (IndicesRELIEF[, 13] / sum(IndicesRELIEF[, 13])) * 100
    IndicesRELIEF[, 16] <- (IndicesRELIEF[, 15] / sum(IndicesRELIEF[, 15])) * 100
    IndicesRELIEF[, 18] <- (IndicesRELIEF[, 17] / sum(IndicesRELIEF[, 17])) * 100

    rm(numVAR, nIndices, w, nOrnessOWA, orness)

    IndicesRELIEF <- as.data.frame(IndicesRELIEF)
    IndicesRELIEF

    varsINPUTselected <- row.names(IndicesRELIEF[order(-IndicesRELIEF$OWApor04), ])[1:3]

    rm(IndicesRELIEF)

    # watchful$txt <- varsINPUTselected
    cat("Self-X Supervised Feature Selection: ", varsINPUTselected, "\n", sep = "\t")
    return(varsINPUTselected)
  } else {
    if (watchful$cod == 1) {
      print(watchful$txt)
      return(0)
    }
  }
} # relief.owa
